/*
 * wifi-signal-applet: Shows the Wi-Fi signal in the status area
 *
 * Copyright (C) 2010 Faheem Pervez <trippin1@gmail.com>. All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *      
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *      
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301, USA.
 *
 */

/* wifi-signal-applet.c */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <icd/network_api.h>
#include <icd/dbus_api.h>
#include <string.h>
#include <glib.h>
#include <glib-object.h>
#include <cairo/cairo.h>
#include <pango/pango.h>
#include <dbus/dbus-glib.h>
#include <conic/conic.h>
#include <gtk/gtk.h>
#include <hildon/hildon.h>
#include <libhildondesktop/libhildondesktop.h>

#include "icd2-stats-sig-marshal.h"

#define UNLIKELY_RETURN(thing) G_STMT_START{ if (G_UNLIKELY (!thing)) return; }G_STMT_END

/* Oh, this one's a fucker. This is one of the few, rare moments where I actually prefer libdbus directly to dbus-glib */
#define ICD2_STATS_SIGNAL_G_TYPES G_TYPE_STRING, G_TYPE_UINT, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_UINT, DBUS_TYPE_G_UCHAR_ARRAY, G_TYPE_UINT, G_TYPE_INT, G_TYPE_UINT, G_TYPE_UINT

/* Start of GObject boilerplate shit */
GType wifi_signal_applet_get_type (void);

#define WIFI_TYPE_SIGNAL_APPLET (wifi_signal_applet_get_type ())

#define WIFI_SIGNAL_APPLET(obj) (G_TYPE_CHECK_INSTANCE_CAST ((obj), \
						WIFI_TYPE_SIGNAL_APPLET, WifiSignalApplet))

#define WIFI_SIGNAL_APPLET_CLASS(klass) (G_TYPE_CHECK_CLASS_CAST ((klass), \
						WIFI_TYPE_SIGNAL_APPLET, WifiSignalAppletClass))

#define WIFI_IS_SIGNAL_APPLET(obj) (G_TYPE_CHECK_INSTANCE_TYPE ((obj), \
						WIFI_TYPE_SIGNAL_APPLET))

#define WIFI_IS_SIGNAL_APPLET_CLASS(klass) (G_TYPE_CHECK_CLASS_TYPE ((klass), \
						WIFI_TYPE_SIGNAL_APPLET))

#define WIFI_SIGNAL_APPLET_GET_CLASS(obj) (G_TYPE_INSTANCE_GET_CLASS ((obj), \
						WIFI_TYPE_SIGNAL_APPLET, WifiSignalAppletClass))

#define WIFI_SIGNAL_APPLET_GET_PRIVATE(obj) (G_TYPE_INSTANCE_GET_PRIVATE ((obj), \
						WIFI_TYPE_SIGNAL_APPLET, WifiSignalAppletPrivate))

typedef struct _WifiSignalApplet WifiSignalApplet;
typedef struct _WifiSignalAppletClass WifiSignalAppletClass;
typedef struct _WifiSignalAppletPrivate WifiSignalAppletPrivate;

struct _WifiSignalApplet
{
	HDStatusMenuItem parent;

	WifiSignalAppletPrivate *priv;
};

struct _WifiSignalAppletClass
{
	HDStatusMenuItemClass parent_class;
};

struct _WifiSignalAppletPrivate
{
	DBusGConnection *sys_dbus_conn;
	DBusGProxy *icd2_proxy;
	ConIcConnection *con;

	gboolean iap_connected;
	gboolean is_status_area_visible;
	guint last_signal_strength;

	guint timeout_src;

	GdkPixbuf *pixbuf;
};

HD_DEFINE_PLUGIN_MODULE (WifiSignalApplet, wifi_signal_applet, HD_TYPE_STATUS_MENU_ITEM)
/* End of GObject boilerplate shit */

static const cairo_user_data_key_t pixbuf_key; 

static void wifi_signal_applet_on_stats_rec_cb (DBusGProxy *object G_GNUC_UNUSED, const gchar *service_type G_GNUC_UNUSED, const guint service_attributes G_GNUC_UNUSED, const gchar *service_id G_GNUC_UNUSED, const gchar *network_type G_GNUC_UNUSED, const guint network_attributes G_GNUC_UNUSED, const GArray *network_id G_GNUC_UNUSED, const guint time_active G_GNUC_UNUSED, const enum icd_nw_levels signal_strength, const guint bytes_sent G_GNUC_UNUSED, const guint bytes_recieved G_GNUC_UNUSED, WifiSignalApplet *plugin);
static void wifi_signal_applet_remove_timeout (WifiSignalApplet *plugin);

/* Thanks to Thomas Perl for the initial versions of these */
static const PangoFontDescription*
he_helper_get_logical_font_desc (const gchar *name)
{
	GtkSettings *settings = gtk_settings_get_default();
	GtkStyle *style = gtk_rc_get_style_by_paths(settings,
				            name, NULL, G_TYPE_NONE);

	return style->font_desc;
}

static gboolean
he_helper_get_logical_font_colour (const gchar *name, guint16 *red, guint16 *green, guint16 *blue)
{
	GdkColor colour;
	GtkSettings *settings = gtk_settings_get_default();
	GtkStyle *style = gtk_rc_get_style_by_paths(settings,
				                "GtkButton", "osso-logical-colors", GTK_TYPE_BUTTON);

	if (gtk_style_lookup_color(style, name, &colour)) {
		*red = colour.red;
		*green = colour.green;
		*blue = colour.blue;

		return TRUE;
	} else {
		return FALSE;
	}
}

/* Thanks to tadeboro: http://www.gtkforums.com/about5204.html */
static cairo_t *
pixbuf_cairo_create( GdkPixbuf *pixbuf )
{
    gint             width,        /* Width of both pixbuf and surface */
                     height,       /* Height of both pixbuf and surface */
                     p_stride,     /* Pixbuf stride value */
                     p_n_channels, /* RGB -> 3, RGBA -> 4 */
                     s_stride;     /* Surface stride value */
    guchar          *p_pixels,     /* Pixbuf's pixel data */
                    *s_pixels;     /* Surface's pixel data */
    cairo_surface_t *surface;      /* Temporary image surface */
    cairo_t         *cr;           /* Final context */

    g_object_ref( G_OBJECT( pixbuf ) );

    /* Inspect input pixbuf and create compatible cairo surface */
    g_object_get( G_OBJECT( pixbuf ), "width",           &width,
                                      "height",          &height,
                                      "rowstride",       &p_stride,
                                      "n-channels",      &p_n_channels,
                                      "pixels",          &p_pixels,
                                      NULL );
    surface = cairo_image_surface_create( CAIRO_FORMAT_ARGB32, width, height );
    s_stride = cairo_image_surface_get_stride( surface );
    s_pixels = cairo_image_surface_get_data( surface );

    /* Copy pixel data from pixbuf to surface */
    while( height-- )
    {
        gint    i;
        guchar *p_iter = p_pixels,
               *s_iter = s_pixels;

        for( i = 0; i < width; i++ )
        {
#if G_BYTE_ORDER == G_LITTLE_ENDIAN
            /* Pixbuf:  RGB(A)
             * Surface: BGRA */
            if( p_n_channels == 3 )
            {
                s_iter[0] = p_iter[2];
                s_iter[1] = p_iter[1];
                s_iter[2] = p_iter[0];
                s_iter[3] = 0xff;
            }
            else /* p_n_channels == 4 */
            {
                gdouble alpha_factor = p_iter[3] / (gdouble)0xff;

                s_iter[0] = (guchar)( p_iter[2] * alpha_factor + .5 );
                s_iter[1] = (guchar)( p_iter[1] * alpha_factor + .5 );
                s_iter[2] = (guchar)( p_iter[0] * alpha_factor + .5 );
                s_iter[3] =           p_iter[3];
            }
#elif G_BYTE_ORDER == G_BIG_ENDIAN
            /* Pixbuf:  RGB(A)
             * Surface: ARGB */
            if( p_n_channels == 3 )
            {
                s_iter[3] = p_iter[2];
                s_iter[2] = p_iter[1];
                s_iter[1] = p_iter[0];
                s_iter[0] = 0xff;
            }
            else /* p_n_channels == 4 */
            {
                gdouble alpha_factor = p_iter[3] / (gdouble)0xff;

                s_iter[3] = (guchar)( p_iter[2] * alpha_factor + .5 );
                s_iter[2] = (guchar)( p_iter[1] * alpha_factor + .5 );
                s_iter[1] = (guchar)( p_iter[0] * alpha_factor + .5 );
                s_iter[0] =           p_iter[3];
            }
#else /* PDP endianness */
            /* Pixbuf:  RGB(A)
             * Surface: RABG */
            if( p_n_channels == 3 )
            {
                s_iter[0] = p_iter[0];
                s_iter[1] = 0xff;
                s_iter[2] = p_iter[2];
                s_iter[3] = p_iter[1];
            }
            else /* p_n_channels == 4 */
            {
                gdouble alpha_factor = p_iter[3] / (gdouble)0xff;

                s_iter[0] = (guchar)( p_iter[0] * alpha_factor + .5 );
                s_iter[1] =           p_iter[3];
                s_iter[1] = (guchar)( p_iter[2] * alpha_factor + .5 );
                s_iter[2] = (guchar)( p_iter[1] * alpha_factor + .5 );
            }
#endif
            s_iter += 4;
            p_iter += p_n_channels;
        }
        s_pixels += s_stride;
        p_pixels += p_stride;
    }

    /* Create context and set user data */
    cr = cairo_create( surface );
    cairo_surface_destroy( surface );
    cairo_set_user_data( cr, &pixbuf_key, pixbuf, g_object_unref );

    /* Return context */
    return( cr );
}

static GdkPixbuf *
pixbuf_cairo_destroy( cairo_t  *cr,
                         gboolean  create_new_pixbuf )
{
    gint             width,        /* Width of both pixbuf and surface */
                     height,       /* Height of both pixbuf and surface */
                     p_stride,     /* Pixbuf stride value */
                     p_n_channels, /* RGB -> 3, RGBA -> 4 */
                     s_stride;     /* Surface stride value */
    guchar          *p_pixels,     /* Pixbuf's pixel data */
                    *s_pixels;     /* Surface's pixel data */
    cairo_surface_t *surface;      /* Temporary image surface */
    GdkPixbuf       *pixbuf,       /* Pixbuf to be returned */
                    *tmp_pix;      /* Temporary storage */

    /* Obtain pixbuf to be returned */
    tmp_pix = cairo_get_user_data( cr, &pixbuf_key );
    if( create_new_pixbuf )
        pixbuf = gdk_pixbuf_copy( tmp_pix );
    else
        pixbuf = g_object_ref( G_OBJECT( tmp_pix ) );

    /* Obtain surface from where pixel values will be copied */
    surface = cairo_get_target( cr );

    /* Inspect pixbuf and surface */
    g_object_get( G_OBJECT( pixbuf ), "width",           &width,
                                      "height",          &height,
                                      "rowstride",       &p_stride,
                                      "n-channels",      &p_n_channels,
                                      "pixels",          &p_pixels,
                                      NULL );
    s_stride = cairo_image_surface_get_stride( surface );
    s_pixels = cairo_image_surface_get_data( surface );

    /* Copy pixel data from surface to pixbuf */
    while( height-- )
    {
        gint    i;
        guchar *p_iter = p_pixels,
               *s_iter = s_pixels;

        for( i = 0; i < width; i++ )
        {
#if G_BYTE_ORDER == G_LITTLE_ENDIAN
            /* Pixbuf:  RGB(A)
             * Surface: BGRA */
            gdouble alpha_factor = (gdouble)0xff / s_iter[3];

            p_iter[0] = (guchar)( s_iter[2] * alpha_factor + .5 );
            p_iter[1] = (guchar)( s_iter[1] * alpha_factor + .5 );
            p_iter[2] = (guchar)( s_iter[0] * alpha_factor + .5 );
            if( p_n_channels == 4 )
                p_iter[3] = s_iter[3];
#elif G_BYTE_ORDER == G_BIG_ENDIAN
            /* Pixbuf:  RGB(A)
             * Surface: ARGB */
            gdouble alpha_factor = (gdouble)0xff / s_iter[0];

            p_iter[0] = (guchar)( s_iter[1] * alpha_factor + .5 );
            p_iter[1] = (guchar)( s_iter[2] * alpha_factor + .5 );
            p_iter[2] = (guchar)( s_iter[3] * alpha_factor + .5 );
            if( p_n_channels == 4 )
                p_iter[3] = s_iter[0];
#else /* PDP endianness */
            /* Pixbuf:  RGB(A)
             * Surface: RABG */
            gdouble alpha_factor = (gdouble)0xff / s_iter[1];

            p_iter[0] = (guchar)( s_iter[0] * alpha_factor + .5 );
            p_iter[1] = (guchar)( s_iter[3] * alpha_factor + .5 );
            p_iter[2] = (guchar)( s_iter[2] * alpha_factor + .5 );
            if( p_n_channels == 4 )
                p_iter[3] = s_iter[1];
#endif
            s_iter += 4;
            p_iter += p_n_channels;
        }
        s_pixels += s_stride;
        p_pixels += p_stride;
    }

    /* Destroy context */
    cairo_destroy( cr );

    /* Return pixbuf */
    return( pixbuf );
} 

static void wifi_signal_applet_update_image (WifiSignalApplet *plugin, gchar *text)
{
    GdkPixbuf   *new_pix;
    PangoLayout *layout;
    cairo_t     *cr;
    guint16 red, green, blue;
    gboolean colour;

    gdk_pixbuf_fill (plugin->priv->pixbuf, 0x00000000);

    cr = pixbuf_cairo_create( plugin->priv->pixbuf );
    cairo_set_source_rgb( cr, 0, 0, 0 );
    cairo_move_to( cr, 0, 0 );
    layout = pango_cairo_create_layout( cr );
    pango_layout_set_font_description (layout, he_helper_get_logical_font_desc ("SmallSystemFont"));
    pango_layout_set_text( layout,
                           text, -1 );
    colour = he_helper_get_logical_font_colour("DefaultTextColor", &red, &green, &blue);
    if (colour)
    	cairo_set_source_rgb(cr, red, green, blue);
    pango_cairo_update_layout(cr, layout);
    pango_cairo_show_layout( cr, layout );
    g_object_unref( G_OBJECT( layout ) );
    new_pix = pixbuf_cairo_destroy( cr, FALSE );
    g_object_unref( G_OBJECT( new_pix ) );

    hd_status_plugin_item_set_status_area_icon (HD_STATUS_PLUGIN_ITEM (plugin), plugin->priv->pixbuf);
}

/* Where it all falls down */
static void wifi_signal_applet_finalize (GObject *object)
{
	WifiSignalApplet *plugin = WIFI_SIGNAL_APPLET (object);

	if (plugin->priv->con)
	{
		g_object_unref (plugin->priv->con);
		plugin->priv->con = NULL;
	}

	wifi_signal_applet_remove_timeout (plugin);

	if (plugin->priv->pixbuf)
	{
		//hildon_banner_show_informationf (NULL, NULL, "%u", G_OBJECT (plugin->priv->pixbuf)->ref_count);
		g_object_unref (plugin->priv->pixbuf);
		plugin->priv->pixbuf = NULL;
	}

	if (plugin->priv->icd2_proxy)
	{
		dbus_g_proxy_disconnect_signal (plugin->priv->icd2_proxy, ICD_DBUS_API_STATISTICS_SIG, G_CALLBACK (wifi_signal_applet_on_stats_rec_cb), plugin);
		g_object_unref (plugin->priv->icd2_proxy);
		plugin->priv->icd2_proxy = NULL;
	}

	/* if (plugin->priv->sys_dbus_conn)
		dbus_g_connection_unref (plugin->priv->sys_dbus_conn); */ /* Allegedly, hd_status_plugin_item_get_dbus_g_connection () returns a shared connection. */

	G_OBJECT_CLASS (wifi_signal_applet_parent_class)->finalize (object);
}

/* -- */

static gboolean wifi_signal_applet_send_stats_req (WifiSignalApplet *plugin)
{
	dbus_g_proxy_call_no_reply (plugin->priv->icd2_proxy, ICD_DBUS_API_STATISTICS_REQ, G_TYPE_INVALID);

	return TRUE;
}

static void wifi_signal_applet_remove_timeout (WifiSignalApplet *plugin)
{
	if (plugin->priv->timeout_src != 0)
		g_source_remove (plugin->priv->timeout_src);
	
	plugin->priv->timeout_src = 0;
}

static void wifi_signal_applet_add_timeout (WifiSignalApplet *plugin)
{
	wifi_signal_applet_remove_timeout (plugin);
	if (plugin->priv->is_status_area_visible && plugin->priv->iap_connected)
		plugin->priv->timeout_src = g_timeout_add_seconds (2, (GSourceFunc) wifi_signal_applet_send_stats_req, plugin);
	else if (!plugin->priv->iap_connected)
	{
		hd_status_plugin_item_set_status_area_icon (HD_STATUS_PLUGIN_ITEM (plugin), NULL);
		plugin->priv->last_signal_strength = 0;
	}
}

static void wifi_signal_applet_on_stats_rec_cb (DBusGProxy *object G_GNUC_UNUSED, const gchar *service_type G_GNUC_UNUSED, const guint service_attributes G_GNUC_UNUSED, const gchar *service_id G_GNUC_UNUSED, const gchar *network_type G_GNUC_UNUSED, const guint network_attributes G_GNUC_UNUSED, const GArray *network_id G_GNUC_UNUSED, const guint time_active G_GNUC_UNUSED, const enum icd_nw_levels signal_strength, const guint bytes_sent G_GNUC_UNUSED, const guint bytes_recieved G_GNUC_UNUSED, WifiSignalApplet *plugin)
{
	if (signal_strength != plugin->priv->last_signal_strength)
	{
		plugin->priv->last_signal_strength = signal_strength;

		gchar signal_strength_str[4]; /* If it's any bigger, we're fucked... not to mention that it wouldn't fit anyway */
		g_snprintf (signal_strength_str, sizeof (signal_strength_str), "%u", signal_strength);

		wifi_signal_applet_update_image (plugin, signal_strength_str);
	}
}

static void wifi_signal_applet_iap_connect_event_cb (ConIcConnection *con G_GNUC_UNUSED, ConIcConnectionEvent *e, WifiSignalApplet *plugin)
{
	plugin->priv->iap_connected = con_ic_connection_event_get_status (e) == CON_IC_STATUS_CONNECTED;
	wifi_signal_applet_add_timeout (plugin);
}

static void wifi_signal_applet_status_area_visible_cb (GObject *object, GParamSpec *pspec, gpointer user_data G_GNUC_UNUSED)
{
	WifiSignalApplet *plugin = WIFI_SIGNAL_APPLET (object);
	g_object_get (object, pspec->name, &plugin->priv->is_status_area_visible, NULL);
	wifi_signal_applet_add_timeout (plugin);
}

static void wifi_signal_applet_init (WifiSignalApplet *plugin)
{
	plugin->priv = WIFI_SIGNAL_APPLET_GET_PRIVATE (plugin);
	UNLIKELY_RETURN (plugin->priv);
	memset (plugin->priv, sizeof (WifiSignalAppletPrivate), 0);

	plugin->priv->sys_dbus_conn = hd_status_plugin_item_get_dbus_g_connection (HD_STATUS_PLUGIN_ITEM (plugin), DBUS_BUS_SYSTEM, NULL);
	UNLIKELY_RETURN (plugin->priv->sys_dbus_conn);

	plugin->priv->icd2_proxy = dbus_g_proxy_new_for_name (plugin->priv->sys_dbus_conn, ICD_DBUS_API_INTERFACE, ICD_DBUS_API_PATH, ICD_DBUS_API_INTERFACE);
	UNLIKELY_RETURN (plugin->priv->icd2_proxy);

	plugin->priv->pixbuf = gdk_pixbuf_new (GDK_COLORSPACE_RGB, TRUE, 8, 18, 18);
	gdk_pixbuf_fill (plugin->priv->pixbuf, 0x00000000);

	g_signal_connect (plugin, "notify::status-area-visible", G_CALLBACK (wifi_signal_applet_status_area_visible_cb), NULL);
	g_object_notify (G_OBJECT (plugin), "status-area-visible");

	dbus_g_object_register_marshaller (icd2_stats_marshal_VOID__STRING_UINT_STRING_STRING_UINT_BOXED_UINT_INT_UINT_UINT, G_TYPE_NONE, ICD2_STATS_SIGNAL_G_TYPES, G_TYPE_INVALID);
	dbus_g_proxy_add_signal (plugin->priv->icd2_proxy, ICD_DBUS_API_STATISTICS_SIG, ICD2_STATS_SIGNAL_G_TYPES, G_TYPE_INVALID);
	dbus_g_proxy_connect_signal (plugin->priv->icd2_proxy, ICD_DBUS_API_STATISTICS_SIG, G_CALLBACK (wifi_signal_applet_on_stats_rec_cb), plugin, NULL);

	plugin->priv->con = con_ic_connection_new ();
	UNLIKELY_RETURN (plugin->priv->con);

	g_signal_connect (plugin->priv->con, "connection-event", G_CALLBACK (wifi_signal_applet_iap_connect_event_cb), plugin); 
	g_object_set (plugin->priv->con, "automatic-connection-events", TRUE, NULL);
}

/* ---- */
static void wifi_signal_applet_class_init (WifiSignalAppletClass *klass)
{
	GObjectClass *object_class = G_OBJECT_CLASS (klass);

	object_class->finalize = (GObjectFinalizeFunc) wifi_signal_applet_finalize;

	g_type_class_add_private (klass, sizeof (WifiSignalAppletPrivate));
}

static void wifi_signal_applet_class_finalize (WifiSignalAppletClass *klass G_GNUC_UNUSED)
{
}

